# -*- coding: utf-8 -*-
"""
Created on Mon Nov 15 23:36:15 2021
This file contains the workflow for similar system transfer learning on PIML
with the simple poly functions as systems
@author: wenzel
"""
from tensorflow import keras
import numpy as np
import matplotlib.pyplot as plt
import scipy.interpolate as inter
from pyDOE import lhs
import vecParam as vb



np.random.seed(123) 
#Somehow this is not working for me as it should.
#Different results are produced each time.
#May be a user error, I haven't investigated it closely.
#It only changes the numbers but not the result.


class FRED:
    lhs_saved = 10000
    anz_param = 124
    anz_b = 4
    anz_inp = anz_param + anz_b
    anz_target = 1
    len_lhs = 128
    x_lhs = lhs(anz_param, len_lhs)
    lhs_save = ''
    results_file = 'results.csv'
    F = None
    R = None
    ED = None
    
    
    def int_2_bin_arr(self,integ, length = None):
        integ = int(integ)
        tmp_str = format(integ, "b")
        bin_arr = [c for c in tmp_str]
        bin_arr = np.array(bin_arr)
        bin_arr = bin_arr.astype(np.float64)
        if length is None:
            return bin_arr
        else:
            if length == 0:
                length = self.anz_b
            while len(bin_arr)<length:
                bin_arr = np.insert(bin_arr,0,0)
            while len(bin_arr) > length:
                bin_arr = np.delete(bin_arr,0)
            return bin_arr
        
    def bin_arr_2_int(self,bin_arr):
        integ = 0
        for bii in bin_arr:
            integ = integ*2
            integ += bii
        return int(integ)
        
    
    def phy_system(self,b=None, inp = None, x_anz = None, gen_new = False, load_n_save = True):
        ## parameter for pre learning
        if b is None:
            if gen_new:
                inp_out = 0
                for ii in range(2):
                    b = self.int_2_bin_arr(ii,0)
                    [inp_tmp,outp_tmp] = self.phy_system(b,inp, x_anz)
                    anz = inp_tmp.shape
                    x_anz = anz[0]
                    b = np.array([b for tmp in range(x_anz)])
                    inp_tmp = np.hstack((b,inp_tmp))
                    if ii == 0:
                        inp_out = inp_tmp
                        outp= outp_tmp
                    else:                    
                        inp_out = np.concatenate((inp_out,inp_tmp))
                        outp = np.concatenate((outp,outp_tmp))
                if load_n_save:
                    inp_phy = np.genfromtxt("variables/inp_phy.txt")
                    out_phy = np.genfromtxt("variables/out_phy.txt")
                    
                    inp_phy = np.vstack((inp_phy,inp_out))
                    outp = np.hstack((out_phy,outp))
                    
                    np.savetxt('variables/inp_phy.txt',inp_phy)
                    np.savetxt('variables/out_phy.txt',outp)
            else:
                inp_out = np.genfromtxt("variables/inp_phy.txt")
                outp = np.genfromtxt("variables/out_phy.txt")
            return [inp_out, outp]
        
        if inp is None: 
            if x_anz is None:
                x_anz = 10000
            inp = np.random.rand(x_anz,self.anz_param)
            outp = np.zeros(x_anz)
        else:
            anz = inp.shape
            x_anz = anz[0]
            outp = np.zeros(x_anz)
            
        b = self.bin_arr_2_int(b)
        if b == 0: 
            
            #the influence of first layer parameters on adhesion between the 3D printer’s glass bed and ABS
            #1 round 21 measurments without initial line width
            #2 round 19 measurments
            
            #nozzle_temp
            nt = np.ones(40)*250 
            nt[0:5] = [235, 240, 245, 250, 255]   
            nt[21:40] = np.ones(19)*255
            nt[31:36] = [235,240,245,250,255]
            
            #bed_temp
            bt = np.ones(40)*100
            bt[5:9] = [90, 100, 110, 120]
            bt[21:40] = np.ones(19)*100
            bt[36:40] = [95,100,105,110]
            
            #first_layer_height
            fh = np.ones(40)*0.25
            fh[9:13] = [0.1, 0.25, 0.33, 0.4]
            fh[21:27] = [0.05,0.1,0.2,0.25,0.33,0.4]
    
            #first_layer_speed
            fs = np.ones(40)*20
            fs[13:17] = [10,20,30,40]
            fs[21:40] = np.ones(19)*30
            fs[27:31] = [10,20,30,40]
            
            
            #flow_rate
            fr = np.ones(40)*100
            fr[17:21] = [80,100,120,140]
            fr[21:40] = np.ones(19)*140
            
            
            force = [10, 10.5, 9.5, 10.5, 11.5,#nozzle_temp
                     4.7, 11, 11.7, 9, #bed_temp
                     9,11, 13.5, 13,#first_layer_height
                     12, 10, 13.5, 12,#first_layer_speed
                     4, 5.5, 8, 12,#flow_rate
                     14,17,20,21,17,17,#first_layer_height
                     21,21,25,21,#first_layer_speed
                     21,21,22,21.5,25,#nozzle_temp
                     8,20,28,18] #bed_temp
            
            #interpo = Rbf(nt,bt,fh,fs,fr,force, function='linear', neighbors=5)
            bed_temp = np.zeros(x_anz)
            first_layer_width = np.zeros(x_anz)
            first_layer_height = np.zeros(x_anz)
            normal_layer_width = np.zeros(x_anz)
            nozzle_temp = np.zeros(x_anz)
            first_layerspeed = np.zeros(x_anz)
            flow_rate = np.zeros(x_anz)
            for ii in range(0,x_anz):
                bed_temp[ii] = inp[ii,8]*(120-90)+90
                first_layer_width[ii] = inp[ii,58]*(0.7-0.4)+0.4
                first_layer_height[ii] = inp[ii,59]*(0.4-0.1)+0.1
                normal_layer_width[ii] = inp[ii,36]*(0.7-0.3)+0.3
                nozzle_temp[ii] = inp[ii,101]*(250-235)+235
                first_layerspeed[ii] = inp[ii,60]*(100-10)+10
                flow_rate[ii] = first_layer_width[ii]*100/normal_layer_width[ii]
            outp = inter.griddata((nt,bt,fh,fs,fr),force, (nozzle_temp,bed_temp,first_layer_height,first_layerspeed,flow_rate), method='linear')
               
            
        
        if b == 1: 
            
            #Optimisation of the Adhesion of Polypropylene-Based Materials during Extrusion-Based Additive Manufacturing
                #1 round 9 measurments only bed temp
                #2 round 8 measurments DOE
                
                #round 2
                #nozzle_temp
                nt = [200,200,230,230,200,200,230,230]
                
                #first_layer_height
                fh = [1.6,1.6,1.6,1.6,2.4,2.4,2.4,2.4]
                
                #flow_rate
                fr = [0.1,0.2,0.1,0.2,0.1,0.2,0.1,0.2]
                
                force = [68,26.6,116.8,45.4,155.5,14.5,155.5,88.4]
                
                outp = np.zeros(x_anz)
                bed_temp = np.zeros(x_anz)
                first_layer_width = np.zeros(x_anz)
                first_layer_height = np.zeros(x_anz)
                normal_layer_width = np.zeros(x_anz)
                nozzle_temp = np.zeros(x_anz)
                first_layerspeed = np.zeros(x_anz)
                flow_rate = np.zeros(x_anz)
                cross_section = np.zeros(x_anz)
                flow_rate = np.zeros(x_anz)
                for ii in range(0,x_anz):
                    bed_temp = inp[ii,8]*(120-90)+90
                    first_layer_width = inp[ii,58]*(0.7-0.4)+0.4
                    first_layer_height = inp[ii,59]*(0.4-0.1)+0.1
                    normal_layer_width = inp[ii,36]*(0.7-0.3)+0.3
                    nozzle_temp = inp[ii,101]*(250-235)+235
                    first_layerspeed = inp[ii,60]*(100-10)+10
                    cross_section = (first_layer_width-first_layer_height)*first_layer_height+np.pi*((first_layer_height/2)**2)
                    flow_rate = cross_section*first_layerspeed
                    
                    
                    outp[ii] = inter.griddata((nt,fh,fr),force, (nozzle_temp,first_layer_height,flow_rate), method='nearest')
               
                    
                    #1 round
                    outp[ii] = outp[ii] -29 + np.interp(bed_temp,
                               (20,30,40,50,60,70,80,90,100),
                               (0,0,7,22,29,39,56,85,60))   #29 50° bed temp for all round 2 measurments
              

        for ii in range(x_anz-1,-1,-1):
            if np.isnan(outp[ii]):
                outp = np.delete(outp,ii)
                inp = np.delete(inp,ii, axis=0)
        outp = outp/200    
        return [inp, outp]
    
    
    def get_model(self):
        "get_model: give back a new untrained modell for transfer learning"
        model = keras.Sequential()
            
        
            
        dropout = 0.15
        
        model.add(keras.layers.Dense(62, activation='tanh'))
        model.add(keras.layers.Dropout(dropout))
        model.add(keras.layers.Dense(31, activation='tanh'))
        model.add(keras.layers.Dropout(dropout))
        model.add(keras.layers.Dense(16, activation='tanh'))
        model.add(keras.layers.Dropout(dropout))
        model.add(keras.layers.Dense(8, activation='tanh'))
        model.add(keras.layers.Dropout(dropout))
        model.add(keras.layers.Dense(4, activation='tanh'))
        model.add(keras.layers.Dropout(dropout))
        model.add(keras.layers.Dense(2, activation='tanh'))
        model.add(keras.layers.Dropout(dropout))
        model.add(keras.layers.Dense(1, activation='tanh'))
        
        model.compile(optimizer='adam',#tf.train.RMSPropOptimizer(0.001),
                       loss=keras.metrics.mean_squared_error)
        
        return model
    
    
    
    def get_ylhs(self,model,b):
        shape_b = np.shape(b)
        if len(shape_b) == 2:
            ylhs = []
            for b_ii in range(shape_b[0]):
                ylhs.append(self.get_ylhs(model,b[b_ii,:]))
            ylhs = np.array(ylhs)
             #ylhs = np.reshape(ylhs,(ylhs.shape[0],ylhs.shape[1]))
        else:
            inp_b = [b for ii in range(self.len_lhs)]
            inp_b = np.array(inp_b)
            inp = np.hstack((inp_b,self.x_lhs))
            ylhs = model.predict(inp)
        return ylhs
    
    def get_autoencoder(self,samples):
        
        self.ED = keras.Sequential()
        
        dropout = 0.15
        
        dimi = 2
        n = np.floor(np.log(self.len_lhs/self.anz_b)/np.log(dimi))
        
        for ii in range(1,n):
            model.add(keras.layers.Dropout(dropout))
            model.add(keras.layers.Dense(np.floor(self.len_lhs/(dimi ** ii)), activation='tanh')) 
            
        model.add(keras.layers.Dense(self.anz_b, activation='tanh'))  
            
        for ii in range(n,-1,-1):
            model.add(keras.layers.Dropout(dropout))
            model.add(keras.layers.Dense(np.floor(self.len_lhs/(dimi ** ii)), activation='tanh'))
            
        
        return 0
    

        
    def get_BV(self,samples):
        
        #get_cfv = keras.Model([self.ED.layers[0].input],
        #                          [self.ED.layers[self.mid_layer].output])

        get_cfv = keras.Model(inputs=self.ED.input,
                                      outputs = self.ED.layers[self.mid_layer].output)
        anz_sampl = samples.shape[0]
        BV = np.zeros((anz_sampl,self.anz_b))
        
        
        for ii in range(anz_sampl):
            BV[ii,:] = get_cfv(samples[ii,:])
        
        return BV
    
    def get_rnn_training_data(self,samples,BVs,anz_seq,len_seq):
        anz_samples = samples.shape[0]
        len_BV = BVs.shape[1]
        anz_train = anz_seq*len_seq*anz_samples
        data_in = np.zeros((anz_train,2))
        data_out = np.zeros((anz_train,len_BV))
        act_ind = 0
        
        sample_inds = np.zeros(anz_samples*anz_seq)
        for ii in range(anz_samples):
            for jj in range(anz_seq):
                sample_inds[act_ind] = ii
                act_ind = ++act_ind
        sample_inds = np.random.permutation(sample_inds)
        act_ind = 0
        for ii in sample_inds:
            sample = samples[ii,:]
            BV = BVs[ii,:]
            data_in[act_ind,:] = np.zeros(1,2)
            data_out[act_ind,:] = np.zeros(1,len_BV)
            act_ind = ++act_ind
            
            rands = np.random.randint(self.len_lhs,size=(len_seq))
            for rand in rands:
                data_in[act_ind,:] = np.concatanate((self.x_lhs[rand],self.sample[rand]), axis=1)
                data_out[act_ind,:] = BV
                act_ind = ++act_ind
                
        return (data_in,data_out)
    
    def get_rnn(self):
        dropout = 0.4
        rnn = keras.Sequential()
        rnn.add(keras.layers.LSTM(4))
        rnn.add(keras.layers.Dropout(dropout))
        rnn.add(keras.layers.LSTM(4))
        rnn.add(keras.layers.Dropout(dropout))
        rnn.add(keras.layers.LSTM(4))
        rnn.add(keras.layers.Dropout(dropout))
        rnn.add(keras.layers.LSTM(5))
        rnn.add(keras.layers.Dropout(dropout))
        rnn.add(keras.layers.LSTM(6))
        rnn.add(keras.layers.Dropout(dropout))
        rnn.add(keras.layers.LSTM(6))
        rnn.add(keras.layers.Dropout(dropout))
        
        rnn.compile(optimizer="adam", loss="mse")
        
        return rnn
           
        
    def get_model_training_data(self,samples,BVs):
        anz_samples = samples.shape[0]
        len_samples = samples.shape[1]
        len_BV = BVs.shape[1]
        data_in = np.zeros((anz_samples*len_samples,len_BV+1))
        data_out = np.zeros((anz_samples*len_samples,1))
        act_ind = 0
        for ii in range(anz_samples):
            for jj in range(len_samples):
                data_in[act_ind] = np.concatanete((BVs[ii],self.x_lhs(jj)), axis=1)
                data_out[act_ind] = samples[ii,jj]
                act_ind += 1
                
        return (data_in,data_out)
        
    
    
          
        
    def draw_plot_model(self,model,lbl = '',linstyl = 'solid'):
        anzP = 1000
        x = np.arange(0,2,2/anzP)
        y_pred = model.predict(x)
        plt.plot(x,y_pred,label=lbl,linestyle = linstyl) 
        
        
    def get_statistical_data(self, load_new = False, save_it = False): #dont use true, raw data is not submitted
        legend = vb.getLegend()
        sel = np.hstack((2,np.arange(7,legend.shape[0])))
        legend = legend[sel,:]
        legend[:,0] = range(len(legend))
        if load_new:
            data = vb.getVectors(self.results_file)
            data[:,2] = data[:,2]/4
            data[:,7] = data[:,7]/10
            data[:,8] = data[:,8]/30
            data[:,9] = data[:,9]/40
            data[:,10] = np.zeros((1273))
            bedadhes = data[:,3]
            bedadhes = np.divide(bedadhes,75)
            inp_data = data[:,sel]
            if save_it:
                np.savetxt('variables/bedadhes.txt',bedadhes)
                np.savetxt('variables/inp_data.txt',inp_data)
        else:
            bedadhes = np.genfromtxt("variables/bedadhes.txt")
            inp_data = np.genfromtxt("variables/inp_data.txt")
        
        return [bedadhes, inp_data, legend]
    

    
        
        
#%%        
a = FRED()


max_anz_train = 1018
aufloes = 14
[bedadhes, inp_data, legend] = a.get_statistical_data()

[inp_phy, out_phy] = a.phy_system()

bin_printer = [a.int_2_bin_arr(c+1,0) for c in inp_data[:,0]]
inp_data_bin = np.hstack((bin_printer,inp_data[:,1:]))
bedadhes_bin = bedadhes
inp_data_test = inp_data[max_anz_train:,:]
bedadhes_test = bedadhes[max_anz_train:]
inp_data = inp_data[:max_anz_train,:]
bedadhes = bedadhes[:max_anz_train]
inp_data_phy_test = inp_data_bin[max_anz_train:,:]
bedadhes_phy_test = bedadhes_test
model = a.get_model()
inp_phy = inp_phy[:3000,:]
out_phy = out_phy[:3000]
inp_data_phy = np.vstack((inp_data_bin,inp_phy))
bedadhes_phy = np.hstack((bedadhes_bin,out_phy))


scores_stat = np.zeros((aufloes,10))
scores_pre = np.zeros((aufloes,10))
anz_train = [np.exp(ii/2)/np.exp((aufloes-1)/2)*max_anz_train for ii in range(aufloes)]
anz_train = np.array(anz_train)
anz_train = anz_train.astype(np.int)


#%%
for jj in range(9):
    for ii in range(aufloes):  
        model = a.get_model()
        if anz_train[ii] == 1:
            val_split = 0
        else:
            val_split = 0.2
        history = model.fit(inp_data[:anz_train[ii],:], bedadhes[:anz_train[ii]], validation_split=val_split,
                  batch_size=128, epochs=50, verbose=0)
        
        score = model.evaluate(inp_data_test, bedadhes_test, verbose = 0)
        scores_stat[ii,0] = anz_train[ii]
        scores_stat[ii,jj+1] = score[0]*75*75
        print(ii+0.5)
        print(scores_stat[ii,1])
        
        model = a.get_model()
    
        history = model.fit(inp_phy, out_phy, validation_split=val_split,
                 batch_size=128, epochs=50, verbose=0)   
        inp_data_phy = inp_data_bin[:anz_train[ii],:]
        bedadhes_phy = bedadhes[:anz_train[ii]]
        
        history = model.fit(inp_data_phy, bedadhes_phy, validation_split=val_split,
                 batch_size=128, epochs=50, verbose=0)   
        score = model.evaluate(inp_data_phy_test, bedadhes_phy_test, verbose = 0)
        scores_pre[ii,0] = anz_train[ii]
        scores_pre[ii,jj+1] = score[0]*75*75
        
        print(ii+1)
        print(scores_pre[ii,1])
#%%

np.savetxt('variables/scores_pre2.txt',scores_pre)
np.savetxt('variables/scores_stat2.txt',scores_stat)

#%%

scores_pre = np.genfromtxt("variables/scores_pre2.txt")
scores_stat = np.genfromtxt("variables/scores_stat2.txt")

scores_mean = np.zeros((14,2))
scores_var = np.zeros((14,2))
for ii in range(14):
    scores_mean[ii,0] = np.sqrt(np.mean(scores_stat[ii,1:10]))
    scores_mean[ii,1] = np.sqrt(np.mean(scores_pre[ii,1:10]))
    
    scores_var[ii,0] = np.var(np.sqrt(scores_stat[ii,1:10]))
    scores_var[ii,1] = np.var(np.sqrt(scores_pre[ii,1:10]))
    
plt.plot(scores_pre[:,0],scores_mean[:,0],label='statistical approach')
plt.plot(scores_pre[:,0],scores_mean[:,1],label='proposed PIML approach', color='xkcd:bright green')


plt.xlabel('number of measurements')
plt.ylabel('RMSE')
plt.xscale('log')
plt.legend(loc='upper right')
plt.grid(True)
plt.savefig('case_study_compare_stat_piml.png',dpi=500)
plt.show()

